
/*
 * map.h - Map data structure
 */

#ifndef _MAP_INCLUDED
#define _MAP_INCLUDED

#include "constants.h"
#include "vector3d.h"
#include <vector>
 
/*
 * Wall types - solid or a corner piece.
 * (or none - obviously)
 */
 
enum WallType {
	WallTypeNone,
	WallTypeSolid,
	// Closed doors.
	WallTypeDoorX,
	WallTypeDoorY,
	// Probably unused.
	WallTypeBottomLeft,
	WallTypeBottomRight,
	WallTypeTopLeft,
	WallTypeTopRight
};

/* Floors are either flat or sloped. The gradient is 1/4 of a
 floor per tile.
 
 This means that once you've gone up 2 tiles, you're half way to
 the next level, at which point it flips you into the next level.
 
 Once you reach the next level, the tile on that floor must be a 
 -2 floor height with the same slope as the previous one (and vice versa).
 
 When a sloped piece is adjacent to a non-sloped, non-solid piece (or a different type),
 a triangle should be drawn.
 
 If I don't have time, I may just design the maps so that this doesn't happen.
*/

enum FloorType {
	FloorTypeFlat,
	FloorTypeWater, // Below the normal level.
	FloorTypeXPositive, // Height increases in +X dir
	FloorTypeXNegative, // Height increases in -X dir
	FloorTypeYPositive, // Height increases in +Y dir
	FloorTypeYNegative, // Height increases in -Y dir
	FloorTypeMissing // Empty gap. You can fall through this, but more
	// Commonly it's used for the tile on the level above the bottom of
	// A slope.
};
 
enum MapSpecial {
	NoSpecial, // Nothing special
	ExitSpecial,
	SecretExitSpecial
};

struct MapBlock {
	unsigned char groundTile;
	unsigned char wallTile;
	WallType wallType;
	FloorType floorType;
	int heightDifference; // Typically -2, -1, 0, or 1
	// If set to false, the block is not rendered in any way.
	// Used for secret areas, and the insides of completely blank bits of map.
	bool visible;
	// special (breakable etc)
	MapSpecial special;
	// Initialise the walls to solid.
	MapBlock() : groundTile(0), wallType(WallTypeSolid),
		floorType(FloorTypeFlat), heightDifference(0), visible(true)  {
	}
};

const int MapSize = 100;
const int MapLayers = 4;

class Manager; // Forward declar.

class Map {
	public:
		
		Manager *manager;
		Map() {
			// Allocate the map data.
			blockMap.resize(MapLayers);
			for (int l=0; l<MapLayers; l++) {
				blockMap[l].resize(MapSize);
				for (int r=0; r<MapSize; r++) {
					blockMap[l][r].resize(MapSize);
				}
			}
			waveTimer = 0;
			runLogic();
		}
		
		MapBlock & blockAt(int x, int y, int z) {
			if ((x<0) || (x >= MapSize) ||
			    (y<0) || (y >= MapSize) ||
			    (z<0) || (z >= MapLayers)) {
				return blockMap[0][0][0];
			}
			return blockMap[z][y][x];
		}
		
		bool isBlocked(int x, int y, int z) {
			MapBlock & b = blockAt(x,y,z);
			return b.wallType != WallTypeNone;
		}
		// Mostly for testing.
		void loadEmptyMap();
		
		bool renderBlock(int x, int y, int z);
		void runLogic();
		
		typedef std::vector<MapBlock> MapRow;
		typedef std::vector<MapRow> MapLayer;
		void load(Manager *manager, const char * filename);
		void makeTileVisible(int x,int y, int z);
		
	private:
		typedef std::vector<MapLayer> WholeMap;
		WholeMap blockMap;
		void renderFlatFloor(int x, int y, int z, const MapBlock & block);
		void renderWaterFloor(int x, int y, int z, 
			const MapBlock & block,  const MapBlock & ne, const MapBlock & nw);
		Component waveTimer;
		Component waterHeightMap[16];
		Vect3d waterNormalMap[16];
		void loadBlock(Manager *manager, char c, int x, int y, int z);
};

#endif // _MAP_INCLUDED


